import hashlib
import inspect

from selenium import webdriver

import common
import log2l


# TODO overwrite SwitchTo
# TODO move step text to resources
class WebDriverBase(common.FindOverride):
    implicitly_wait_timeout = 0
    script_wait_timeout = 0
    page_load_timeout = 0

    def implicitly_wait(self, time_to_wait):
        tm = float(time_to_wait)
        old = common.WAIT_ELEMENT_TIMEOUT
        common.WAIT_ELEMENT_TIMEOUT = tm
        self.implicitly_wait_timeout = tm
        return old

    def set_script_timeout(self, time_to_wait):
        # noinspection PyUnresolvedReferences
        super(WebDriverBase, self).set_script_timeout(time_to_wait)
        old = self.script_wait_timeout
        self.script_wait_timeout = time_to_wait
        return old

    def set_page_load_timeout(self, time_to_wait):
        # noinspection PyUnresolvedReferences
        super(WebDriverBase, self).set_page_load_timeout(time_to_wait)
        old = self.page_load_timeout
        self.page_load_timeout = time_to_wait
        return old

    @property
    def driver(self):
        return self

    @log2l.step
    def get(self, url):
        # noinspection PyUnresolvedReferences
        super(WebDriverBase, self).get(url)

    @log2l.step('Navigate one step backward in the browser history.')
    def back(self):
        # noinspection PyUnresolvedReferences
        super(WebDriverBase, self).back()

    @log2l.step('Navigate one step forward in the browser history.')
    def forward(self):
        # noinspection PyUnresolvedReferences
        super(WebDriverBase, self).forward()

    @log2l.step('Refresh the current page.')
    def refresh(self):
        # noinspection PyUnresolvedReferences
        super(WebDriverBase, self).refresh()

    def __hash__(self):
        # noinspection PyUnresolvedReferences
        return int(hashlib.md5(self.session_id).hexdigest(), 16)

    def __eq__(self, other):
        # noinspection PyUnresolvedReferences
        return other is not None and hasattr(other, "session_id") and self.session_id == other.session_id


class ChromeDriver(WebDriverBase, webdriver.Chrome):
    pass


class BlackBerryDriver(WebDriverBase, webdriver.BlackBerry):
    pass


class AndroidDriver(WebDriverBase, webdriver.Android):
    pass


class FirefoxDriver(WebDriverBase, webdriver.Firefox):
    pass


class SafariDriver(WebDriverBase, webdriver.Safari):
    pass


class RemoteDriver(WebDriverBase, webdriver.Remote):
    pass


class PhantomJSDriver(WebDriverBase, webdriver.PhantomJS):
    pass


class OperaDriver(WebDriverBase, webdriver.Opera):
    pass


class EdgeDriver(WebDriverBase, webdriver.Edge):
    pass


class IeDriver(WebDriverBase, webdriver.Ie):
    pass


BROWSER_MAPPING = {"remote": RemoteDriver,
                   "phantomjs": PhantomJSDriver,
                   "chrome": ChromeDriver,
                   "opera": OperaDriver,
                   "edge": EdgeDriver,
                   "blackberry": BlackBerryDriver,
                   "safari": SafariDriver,
                   "firefox": FirefoxDriver,
                   "android": AndroidDriver,
                   "ie": IeDriver}


def get_driver(browser='firefox', args=None):
    """
    :param browser:
    :param args:
    :rtype: RemoteDriver
    :return:
    """
    if browser not in BROWSER_MAPPING.keys():
        raise RuntimeError("unknown browser %s. allowed: %s" % (browser, ", ".join(BROWSER_MAPPING.keys())))
    driver_cls = BROWSER_MAPPING.get(browser)
    safe_args = {}
    if args is not None:
        expected_arguments = inspect.getargspec(driver_cls.__init__).args
        expected_arguments.remove("self")
        for arg in expected_arguments:
            if arg in args:
                safe_args[arg] = args[arg]
    return driver_cls(**safe_args)



from .core import log2l
from .core.action_chains import ActionChains
from .core.common import PageElementsContainer
from .core.elements import PageElement
from .core.elements import PageElementsList
from .core.webdrivers import get_driver
from .core import waiter



import functools
import hashlib
import re
import time
import uuid

from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException, \
    ElementNotVisibleException
from selenium.webdriver.remote.webelement import WebElement

import common
import log2l
from .waiter import wait_displayed, wait


def need_interaction(func):
    @functools.wraps(func)
    def wrap(*args, **kwargs):
        el = args[0]
        el._wait_ready_for_interaction = True
        try:
            return func(*args, **kwargs)
        finally:
            el._wait_ready_for_interaction = False

    return wrap


class PageElement(common.BasePageElement, common.PageElementsContainer, common.FindOverride, WebElement):
    """ Base element for 'page object' pattern.
    Encapsulate logic of changing of context of element usage.

    Catch StaleElementReferenceException, if it occurs, and try to find the element again,
    if element found then continue execution, otherwise raise NoSuchElementException

    Class instance may be used as a part of page object/page block
    or class type may be specified directly as last parameter of find/child_element(s) to wrap found element(s).

    For page object field may be specified __set__ method for more useful interaction with the element.
    In this case ``_fill_owner`` should be called first in the __set__ method.

    Examples:

    class SomePageBlock(PageElement):
        field = PageElement("#field_id")
        element = PageElement("#element_id")

        def do_some_work(self, keys):
            self.field.send_keys(keys)
            self.element.click()
            ....

        def __set__(self, owner, value):
            self._fill_owner(owner)
            self.do_some_work(value)


    class SomePageObject(object):
        element = SomePageBlock("#block_id")

        def __init__(self, driver):
            self.driver = driver

    SomePageObject(driver).element.do_some_work("bla-bla")
    # it equals to
    SomePageObject(driver).element = "bla-bla"

    driver.child_element("another block selector", SomePageBlock).do_some_work("cha-cha-cha")
    # but the following wont work
    driver.child_element("another block selector", SomePageBlock) = "cha-cha-cha"

    """

    def __init__(self, selector, timeout=None, name=None):
        super(PageElement, self).__init__(selector, name, timeout)
        self._parent = None
        self._id = None
        self.__cache = {}
        self._wait_ready_for_interaction = False

    def has_class(self, class_name):
        cls_attr = self.get_attribute('class')
        return False if cls_attr is None else re.search(r'(^|\s){}(\s|$)'.format(class_name), cls_attr) is not None

    def exists(self):
        """
        :return: True if element is present in the DOM, otherwise False.
                Ignore implicit and element timeouts and execute immediately.
        """
        t = self.wait_timeout
        self.wait_timeout = 0
        try:
            self.reload()
            return True
        except NoSuchElementException:
            return False
        finally:
            self.wait_timeout = t

    def is_displayed(self):
        """
        :return: False if element is not present in the DOM or invisible, otherwise True.
                Ignore implicit and element timeouts and execute immediately.

        To wait when element displayed or not, use ``waiter.wait_displayed`` or ``waiter.wait_not_displayed``
        """
        t = self.wait_timeout
        self.wait_timeout = 0
        try:
            return super(PageElement, self).is_displayed()
        except NoSuchElementException:
            return False
        finally:
            self.wait_timeout = t

    @property
    def id(self):
        if self._id is None:
            self.reload()
        return self._id

    @log2l.step
    def send_keys(self, *value):
        super(PageElement, self).send_keys(*value)

    @log2l.step
    def submit(self):
        super(PageElement, self).submit()

    @log2l.step
    @need_interaction
    def click(self):
        super(PageElement, self).click()

    @log2l.step
    @need_interaction
    def clear(self):
        super(PageElement, self).clear()

    def reload(self):
        we = wait(common.find, self.wait_timeout, owner=self._owner, locator=self._locator)
        if not we:
            raise NoSuchElementException("Element with selector {} was not found".format(self._locator))
        self._id = we.id
        self._parent = we.parent
        self.__cache[self._owner] = self._id

    def _fill_owner(self, owner):
        super(PageElement, self)._fill_owner(owner)
        if self.__cached__ and self._id is not None:
            self._id = self.__cache.get(self._owner)

    def _execute(self, command, params=None):
        if not self.__cached__ or self._id is None:
            self.reload()

        execute_attempts = 0
        while True:
            try:
                if self._wait_ready_for_interaction:
                    self._wait_ready_for_interaction = False
                    if not wait_displayed(self):
                        raise ElementNotVisibleException("Element with selector {}".format(self._locator))
                    self._wait_ready_for_interaction = True
                val = super(PageElement, self)._execute(command, params)
                return val
            except StaleElementReferenceException:
                if execute_attempts > common.WAIT_STALE_ELEMENT_MAX_TRY:
                    raise
                time.sleep(common.WAIT_ELEMENT_POLL_FREQUENCY)
                self.reload()
            execute_attempts += 1
        return None

    def __hash__(self):
        return int(hashlib.md5(self.id).hexdigest(), 16)


class _ListItem(object):
    def __init__(self, container, index):
        """
        :type container: PageElementsList
        :param container:
        :param index:
        """
        # noinspection PyArgumentList
        super(_ListItem, self).__init__("no_selector")
        self._container = container
        self._index = index
        self.__cached__ = True
        self._name = "{0}[{1}]".format(container.name, index)

    def reload(self):
        self._container.reload()
        if self._index >= len(self._container):
            raise NoSuchElementException("Element #%i no longer exists in "
                                         "the '%s'" % (self._index, self._container.name))

    @property
    def wait_timeout(self):
        return self._container.wait_timeout

    @wait_timeout.setter
    def wait_timeout(self, t):
        self._container.wait_timeout = t


class PageElementsList(common.BasePageElement):
    """
    Provide list interface for list of page elements.
    The list encapsulate logic of changing of context of elements usage and hold context of element usage.

     Example:

     class Page(SomeBasePageClass):
        table_rows = PageElementsList('tr')

     assert len(Page().table_rows) == 10  # we expect 10 rows
     assert Page().table_rows.is_displayed()  # at least one row is displayed
     assert Page().table_rows[1].is_displayed()  # second row is displayed
     Page().table_rows[0].find_elements('td')[2].click()  # click on third cell of first row

    """

    def __init__(self, selector, el_class=PageElement, timeout=None, name=None):
        """
        :type selector: tuple[str, str]|str
        :param selector:
        :type el_class: T <= PageElement
        :param el_class:
        :type name: str
        :param name:
        :return:
        """
        super(PageElementsList, self).__init__(selector, name, timeout)
        self._el_class = type("ListOf" + el_class.__name__ + uuid.uuid4().get_hex(), (_ListItem, el_class,), {})
        self.__cache = {}
        self.__items = []

    def is_displayed(self):
        """
        :return: True id at least one element is displayed, otherwise False.
                Ignore implicit and element timeouts and execute immediately.
        """
        t = self.wait_timeout
        self.wait_timeout = 0
        try:
            self.reload()
            return any(e.is_displayed() for e in self)
        finally:
            self.wait_timeout = t

    def reload(self):
        # noinspection PyUnresolvedReferences
        # noinspection PySuperArguments
        l = wait(lambda: super(common.FindOverride, self._owner).find_elements(*self._locator), self.wait_timeout)
        cache = [w.id for w in l]
        self.__initialize_elements(cache)
        self.__cache[self._owner] = cache

    def __initialize_elements(self, items):
        new_len = len(items)
        old_len = len(self.__items)
        if new_len > old_len:
            self.__items.extend([self._el_class(self, i) for i in range(old_len, new_len)])
        else:
            del self.__items[new_len:]
        for e, l in zip(self.__items, items):
            setattr(e, "_id", l), setattr(e, "_parent", self._parent)
            setattr(e, "_w3c", self._w3c), setattr(e, "_owner", self._owner)

    def _fill_owner(self, owner):
        super(PageElementsList, self)._fill_owner(owner)
        if self.__cached__ and self._owner in self.__cache:
            self.__initialize_elements(self.__cache[self._owner])

    def __len__(self):
        self.reload()
        return len(self.__items)

    def __getitem__(self, index):
        if not self.__cached__ or self._owner not in self.__cache:
            self.reload()
        try:
            return self.__items[index]
        except IndexError:
            self.reload()
            return self.__items[index]

    def __iter__(self):
        self.reload()
        i = 0
        while True:
            try:
                yield self[i]
                i += 1
            except IndexError:
                return


class VirtualElement(common.BasePageElement, common.PageElementsContainer, common.FindOverride):
    # TODO add parameters to support parametrized selectors for child elements
    def __init__(self, name=None):
        super(VirtualElement, self).__init__(("virtual", "element"), name)

    def _init_element(self, element):
        # noinspection PyProtectedMember
        element._fill_owner(self._owner)
        return element

    def reload(self):
        raise TypeError("VirtualElementsGroup is not supposed to be reloaded, "
                        "it doesn't have corresponding DOM element ")



import time

from selenium.common.exceptions import TimeoutException

import common


class Waiter(object):
    """
    call ``method(**kwargs)`` at least once, if ``condition(value)`` returns not true
    wait until ``condition(value)`` returns true or ``timeout``.

    Return result of last ``method`` call or rise ``TimeoutException(fail_on_timeout)``
    if fail_on_timeout is not None and time expired

    Example:
        print Waiter(lambda x: x<0).start(lambda: 4, 0)  # immediately print 4

        Waiter(lambda x: x<0)(lambda: 4, 2, "fail")  # after 2 second waiting raise TimeoutException

    """

    def __init__(self, condition):
        self.__condition = condition

    def start(self, method, timeout, fail_on_timeout=None, **kwargs):
        end_time = time.time() + timeout
        value = method(**kwargs)
        check = self.__condition(value)
        while time.time() < end_time and not check:
            time.sleep(common.WAIT_ELEMENT_POLL_FREQUENCY)
            value = method(**kwargs)
            check = self.__condition(value)
        else:
            if not check and fail_on_timeout is not None:
                raise TimeoutException(fail_on_timeout)
        return value

    def __call__(self, method, timeout, fail_on_timeout=None, **kwargs):
        return self.start(method, timeout, fail_on_timeout, **kwargs)


def wait(method, timeout, fail_on_timeout=None, **kwargs):
    """
    Wait ``timeout`` seconds until ``method(**kwargs)`` returns a ``value`` that *bool(value)==True*.
    Returns last ``value``.
     If time expired and ``fail_on_timeout`` specified, then raise TimeoutException.

    :param method:
    :param timeout:
    :param fail_on_timeout:
    :param kwargs:
    :return:
    """
    return Waiter(lambda value: bool(value)).start(method, timeout, fail_on_timeout, **kwargs)


def wait_not(method, timeout, fail_on_timeout=None, **kwargs):
    """
    Wait ``timeout`` seconds until ``method(**kwargs)`` returns a ``value`` that *not value==True*.
    Returns last ``value``.
     If time expired and ``fail_on_timeout`` specified, then raise TimeoutException.

    :param method:
    :param timeout:
    :param fail_on_timeout:
    :param kwargs:
    :return:
    """
    return Waiter(lambda value: not value).start(method, timeout, fail_on_timeout, **kwargs)


def wait_displayed(element, timeout=None, fail_on_timeout=None):
    """
    Wait until element becomes visible or time out.
    Returns true is element became visible, otherwise false.
    If timeout is not specified or 0, then uses specific element wait timeout.
    :param element:
    :param timeout:
    :param fail_on_timeout:
    :return:
    """
    return wait(lambda: element.is_displayed(), timeout or element.wait_timeout, fail_on_timeout)


def wait_not_displayed(element, timeout=None, fail_on_timeout=None):
    """
    Wait until element becomes invisible or time out.
    Returns true is element became invisible, otherwise false.
    If timeout is not specified or 0, then uses specific element wait timeout.
    :param element:
    :param timeout:
    :param fail_on_timeout:
    :return:
    """
    return wait(lambda: not element.is_displayed(), timeout or element.wait_timeout, fail_on_timeout)


# noinspection PyPep8Naming
class skip_implicit_wait(object):
    def __init__(self, element, *elements):
        els = [element]
        els.extend(elements)
        self._timeouts = [(e, e.wait_timeout) for e in els]

    def __enter__(self):
        for e, _ in self._timeouts:
            e.wait_timeout = 0

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        for e, t in self._timeouts:
            e.wait_timeout = t



from distutils.core import setup

long_description = """
pypo4sel.core
--------------------------------------------------------

Wrapper for selenium webdriver to make page objects easy::

    class SomePageBlock(PageElement):
        filed = PageElement("#filed_id", timeout=10)
        button = PageElement("//path/to/element")

        def do_some_work(self, keys):
            self.field.send_keys(keys)
            self.button.click()
            ...

    class SomePageObject(PageElementsContainer):
        element = SomePageBlock(".block_class")

        def __init__(self, driver):
            self.driver = driver

    page = SomePageObject(get_driver('firefox'))
    assert page.element.button.is_displayed()
    page.element.do_some_work("bla-bla")


- lazy element loading by request
- automated handling of `StaleElementReferenceException`
- flexible timeouts
- automated detecting of locator type
- smart lists of elements, automated logs and much more `here <http://github.com/aksas/pypo4sel>`_.
"""

setup(
    name='pypo4sel.core',
    version='0.0.2',
    packages=['pypo4sel', 'pypo4sel.core'],
    url='https://github.com/aksas/pypo4sel',
    license='MIT',
    author='Oleksii Skliarov',
    author_email='oleksii.skliarov@gmail.com',
    description='page object wrapper for selenium webdriver',
    long_description=long_description,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
    ],
    keywords=['selenium', 'webdriver', 'automated', 'testing', 'page object'],
    install_requires=['six', 'selenium'],
)



import time

from selenium import webdriver
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.remote.command import Command

from .common import WAIT_STALE_ELEMENT_MAX_TRY, WAIT_ELEMENT_POLL_FREQUENCY


class ActionChains(webdriver.ActionChains):
    def __move_to(self, element, xoffset=None, yoffset=None):
        params = {'element': element.id}
        if xoffset is not None and yoffset is not None:
            params.update({'xoffset': int(xoffset),
                           'yoffset': int(yoffset)})

        def loop():
            attempt = 0
            while attempt < WAIT_STALE_ELEMENT_MAX_TRY:
                try:
                    self._driver.execute(Command.MOVE_TO, params)
                    break
                except StaleElementReferenceException:
                    time.sleep(WAIT_ELEMENT_POLL_FREQUENCY)
                    element.reload()
                    params['element'] = element.id
                attempt += 1

        return loop

    def move_to_element_with_offset(self, to_element, xoffset, yoffset):
        self._actions.append(self.__move_to(to_element, xoffset, yoffset))
        return self

    def move_to_element(self, to_element):
        self._actions.append(self.__move_to(to_element))
        return self



"""

class Element(PageElement):

    @logging.step
    def method(self, *fargs, **fkwargs):
        do_smth_here

class Page(object):
    my_element = Element('#id')


each time when 'method' will be called the 'start_step' will be called with the following 'options':
 {
    'element_name':'my_element',
    'step_name':'method',
    'args':<fargs>,
    'kwargs':<fkwargs>
 }

if step wrapper will be used like this:
    @logging.step("any text")
    def method(self, *fargs, **fkwargs):

then 'options' parameter of 'start_step' will look like:
 {
    'first_param':'any text'
    'element_name':'my_element',
    'step_name':'method',
    'args':<fargs>,
    'kwargs':<fkwargs>
 }

for example, it may be used for specifying of custom log message format, like
    @logging.step("{step_name} of {my_element} was called")
    def method(self, *fargs, **fkwargs):
        ...

    class MyL(ListenerMixin):
        def start_step(self, step_id, options)
            if 'first_param' in options:
                print options['first_param'].format(**options)
            else:
                print '{my_element}.{step_name}({args}, {fkwargs})'.format(**options)
"""

import functools
import sys
import uuid

__All__ = ["step", "listeners", "message", "action", "debug"]


class Options(object):
    STEP_MESSAGE = "first_param"
    STEP_NAME = "step_name"
    ELEMENT_NAME = "element_name"
    KWARGS = "kwargs"
    ARGS = "args"
    SUPPRESS_CHILD_LOGS = "suppress_child_logs"
    DEBUG_MESSAGE = "debug_message"


class ListenerMixin(object):
    def start_step(self, step_id, **options):
        pass

    def end_step(self, step_id, **options):
        pass

    def exception(self, step_id, err, **options):
        pass

    def message(self, msg, **kwargs):
        pass


listeners = []
""" :type: list[ListenerMixin] """


# noinspection PyPep8Naming
class step(object):
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            return _decorator(args[0], {})
        return super(step, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise Exception("Only one unnamed parameter is allowed.")
        self.options = kwargs
        if len(args) == 1:
            self.options[Options.STEP_MESSAGE] = args[0]

    def __call__(self, method):
        return _decorator(method, self.options)

    def __enter__(self):
        assert Options.STEP_MESSAGE in self.options
        self.step_id = _notify_start(**self.options)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            _notify_exception(self.step_id, (exc_type, exc_val, exc_tb))
        _notify_end(self.step_id)


# noinspection PyPep8Naming
class action(step):
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            return _decorator(args[0], {Options.SUPPRESS_CHILD_LOGS: True})
        return super(step, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        super(action, self).__init__(*args, **kwargs)
        self.options[Options.SUPPRESS_CHILD_LOGS] = True


def debug(msg, **kwargs):
    kwargs.update({Options.DEBUG_MESSAGE: True})
    message(msg, **kwargs)


def attach(msg, contents, *args, **kwargs):
    message(msg, attach=dict(contents=contents, args=args, kwargs=kwargs))


def message(msg, **kwargs):
    [l.message(msg, **kwargs) for l in listeners]


def _decorator(method, step_options):
    step_options.setdefault(Options.STEP_NAME, method.func_name)

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        options = step_options.copy()
        fargs = args
        if len(args) > 0 and hasattr(args[0], "_name"):
            # noinspection PyProtectedMember
            options.setdefault(Options.ELEMENT_NAME, args[0]._name)
        options.setdefault(Options.KWARGS, kwargs)
        options.setdefault(Options.ARGS, args)
        step_id = _notify_start(**options)
        try:
            return method(*fargs, **kwargs)
        except Exception:
            _notify_exception(step_id, sys.exc_info())
            raise
        finally:
            _notify_end(step_id)

    return wrapper


def _notify_start(**options):
    step_id = uuid.uuid4()
    [l.start_step(step_id, **options) for l in listeners]
    return step_id


def _notify_end(step_id, **options):
    [l.end_step(step_id, **options) for l in listeners]


def _notify_exception(step_id, err, **options):
    [l.exception(step_id, err, **options) for l in listeners]



import inspect
import re
from abc import abstractmethod

import six
from selenium.common.exceptions import InvalidSelectorException, NoSuchElementException
from selenium.webdriver.common.by import By

WAIT_STALE_ELEMENT_MAX_TRY = 5
WAIT_ELEMENT_TIMEOUT = 0
WAIT_ELEMENT_POLL_FREQUENCY = 0.5


def get_members_safety(cls):
    # inspect.getmembers calls __get__ method of the field, if exists, that may cause unexpected actions
    # the solution below does't have this problem
    return reduce(lambda a, b: dict(a, **vars(b)), reversed(inspect.getmro(cls)), {}).items()


class PageElementsContainer(object):
    """
    Classes inheriting PageElementsContainer can use BasePageElement(s) as class attributes
    with implicit element initialization and searching.
    If a class inherits PageElementsContainer and doesn't inherit PageElement,
    than it should have attribute 'driver' with web driver instance.
    """

    def __new__(cls, *args, **kwargs):
        for k, v in get_members_safety(cls):
            if isinstance(v, (BasePageElement,)) and v._name is None:
                v._name = k
        # noinspection PyArgumentList
        return super(PageElementsContainer, cls).__new__(cls, *args, **kwargs)

    def all_elements(self):
        """returns all public BasePageElements grouped by this element and it parent(s)
        :rtype: list[(str, BasePageElement)]
        """
        return [(k, getattr(self, k)) for k, v in get_members_safety(self.__class__)
                if not k.startswith("_") and isinstance(v, (BasePageElement,))]


class BasePageElement(object):
    """
    Base class to describe page object element.
    """

    def __init__(self, selector, name=None, timeout=None, cached=True):
        self.__cached__ = cached
        self._locator = build_locator(selector)
        self._name = name
        self._owner = None
        self._parent = None
        """ :type: FindOverride """
        self.__timeout = timeout
        self._w3c = False

    def _fill_owner(self, owner):
        # _parent and _id field are native for Selenium WebDriver WebElement
        # and should be used in this way for clear and through work with PageElement as WebElement
        if hasattr(owner, "parent"):
            self._parent = owner.parent
            self._owner = owner
        else:
            if not hasattr(owner, "driver"):
                raise TypeError("class {0} doesn't have 'driver' attribute.\n"
                                "Class uses page element(s) should inherit "
                                "PageElement or has 'driver' attribute.".format(type(owner).__name__))
            self._parent = owner.driver
            self._owner = owner.driver
        self._w3c = getattr(self._parent, "w3c", False)

    # noinspection PyUnusedLocal
    def __get__(self, owner, cls):
        self._fill_owner(owner)
        return self

    @abstractmethod
    def reload(self):
        pass

    @property
    def wait_timeout(self):
        return WAIT_ELEMENT_TIMEOUT if self.__timeout is None else self.__timeout

    @wait_timeout.setter
    def wait_timeout(self, value):
        self.__timeout = value if isinstance(value, (float, int)) and value >= 0 else None

    @property
    def name(self):
        return self._name or "({}:{})".format(*self._locator)


def define_selector(by, value, el_class):
    """
    :param by:
    :param value:
    :param el_class:
    :rtype: tuple[type, str|tuple[str, str]]
    :return:
    """
    el = el_class
    selector = by
    if isinstance(value, six.string_types):
        selector = (by, value)
    elif value is not None:
        el = value

    if el is None:
        el = elements.PageElement

    return el, selector


class FindOverride(object):
    def child_element(self, by=By.ID, value=None, el_class=None):
        """
        Doesn't rise NoSuchElementException in case if there are no element with the selector.
        In this case ``exists()`` and ``is_displayed()`` methods of the element will return *False*.
        Attempt to call any other method supposed to interact with browser will raise NoSuchElementException.

        usages with ``'one string'`` selector:
         - find_element(by: str) -> PageElement
         - find_element(by: str, value: T <= PageElement) -> T

        usages with ``'webdriver'`` By selector
         - find_element(by: str, value: str) -> PageElement
         - find_element(by: str, value: str, el_class: T <= PageElement) -> T

        :type by: str
        :param by:
        :type value: str | T <= PageElement
        :param value:
        :type el_class: T <= PageElement
        :param el_class:
        :rtype: T <= PageElement
        :return:
        """
        el, selector = define_selector(by, value, el_class)
        return self._init_element(el(selector))

    def child_elements(self, by=By.ID, value=None, el_class=None):
        """
        alias for ``find_elements``
        :param by:
        :param value:
        :param el_class:
        :return:
        """
        el, selector = define_selector(by, value, el_class)
        return self._init_element(elements.PageElementsList(selector, el))

    def find_element(self, by=By.ID, value=None, el_class=None):
        """
        usages with ``'one string'`` selector:
         - find_element(by: str) -> PageElement
         - find_element(by: str, value: T <= PageElement) -> T

        usages with ``'webdriver'`` By selector
         - find_element(by: str, value: str) -> PageElement
         - find_element(by: str, value: str, el_class: T <= PageElement) -> T

        :type by: str
        :param by:
        :type value: str | T <= PageElement
        :param value:
        :type el_class: T <= PageElement
        :param el_class:
        :rtype: T <= PageElement
        :return:
        """
        el = self.child_element(by, value, el_class)
        el.reload()
        return el

    def find_elements(self, by=By.ID, value=None, el_class=None):
        """
        usages with ``'one string'`` selector:
         - find_elements(by: str) -> PageElementsList[ListElement]
         - find_elements(by: str, value: T <= ListElement) -> PageElementsList[T]

        usages with ``'webdriver'`` By selector
         - find_elements(by: str, value: str) -> PageElementsList[ListElement]
         - find_elements(by: str, value: str, el_class: T <= ListElement) -> PageElementsList[T]

        :type by: str
        :param by:
        :type value: str | T <= ListElement
        :param value:
        :type el_class: T <= ListElement
        :param el_class:
        :rtype: PageElementsList[T | ListElement]
        :return:
        """
        els = self.child_elements(by, value, el_class)
        els.reload()
        return els

    def _init_element(self, element):
        # noinspection PyProtectedMember
        element._fill_owner(self)
        return element


import elements


def _stats_with(prefix):
    return lambda s: s.startswith(prefix)


selectors = [
    (re.compile(r"^\w+$").match, By.TAG_NAME, None),
    (re.compile(r"^\.-?[_a-zA-Z]+[_a-zA-Z0-9-]*$").match, By.CLASS_NAME, 1),
    (re.compile(r"^#[A-Za-z]+[:._a-zA-Z0-9-]*$").match, By.ID, 1),
    (re.compile(r"^@[A-Za-z]+[:._a-zA-Z0-9-]*$").match, By.NAME, 1),
    (re.compile(r"^(\./|//).+").match, By.XPATH, None),
    (_stats_with("$x:"), By.XPATH, 3),
    (_stats_with("$link_text:"), By.LINK_TEXT, 11),
    (_stats_with("$partial_link_text:"), By.PARTIAL_LINK_TEXT, 19),
    (re.compile(r"^(\*|\.|#|[\w-]|\[|:).*").match, By.CSS_SELECTOR, None)
]


def build_locator(selector):
    """
    - ID = "#valid_id"
    - CLASS_NAME = ".valid_class_name"
    - TAG_NAME = "valid_tag_name"
    - XPATH = start with "./" or "//" or "$x:"
    - LINK_TEXT = start with "$link_text:"
    - PARTIAL_LINK_TEXT = start with "$partial_link_text:"
    - NAME = "@valid_name_attribute_value"

    CSS_SELECTOR = all other that starts with *|.|#|[\w-]|\[|:

    :type selector: str|tuple
    :param selector:
    :rtype: tuple[selenium.webdriver.common.by.By, str]
    :return:
    """
    if type(selector) is tuple:
        return selector
    if not isinstance(selector, six.string_types):
        raise InvalidSelectorException("Invalid locator values passed in")

    s = selector.strip()
    for test, by, index in selectors:
        if test(s):
            return by, s[index:]
    raise InvalidSelectorException("Invalid locator values passed in: {}".format(selector))


def find(owner, locator):
    try:
        return super(FindOverride, owner).find_element(*locator)
    except NoSuchElementException:
        return False
